{-# LANGUAGE ForeignFunctionInterface, EmptyDataDecls #-}

import Control.Applicative
import Control.Monad.State
import Control.Arrow
import Foreign.C.String
import Foreign.C.Types
import Foreign.Ptr
import System.IO
import Data.Array (Array(..), (!))
import qualified Data.Array as A
import Data.Complex (Complex(..))
import qualified Data.Complex as Complex
import Data.IORef
import Data.List
import Data.Maybe
import Data.Ord
import System.Random
import System.Time
import Debug.Trace


type TouchesFunc = CDouble -> CDouble -> IO ()

foreign import ccall safe "wrapper" mkPlain :: IO () -> IO (FunPtr (IO ()))

foreign import ccall safe "wrapper" mkTouch :: TouchesFunc -> IO (FunPtr TouchesFunc)

foreign import ccall safe "applicationMain" applicationMain_
    :: FunPtr (IO ()) -> FunPtr (IO ()) -> IO CInt

oneShot :: (IO a -> IO (FunPtr (IO a))) -> IO a -> IO (FunPtr (IO a))
oneShot mkFunPtr code = do
    ref <- newIORef Nothing
    ptr <- mkFunPtr $ do
        ptr <- fromJust <$> readIORef ref
        freeHaskellFunPtr ptr
        code
    writeIORef ref (Just ptr)
    return ptr

foreign import ccall safe "setTouchesBegan" setTouchesBegan_ :: FunPtr TouchesFunc -> IO ()
foreign import ccall safe "setTouchesMoved" setTouchesMoved_ :: FunPtr TouchesFunc -> IO ()
foreign import ccall safe "setTouchesEnded" setTouchesEnded_ :: FunPtr TouchesFunc -> IO ()

setTouches_ :: (FunPtr TouchesFunc -> IO ()) -> (Double -> Double -> IO ()) -> IO ()
setTouches_ set code = do
    ptr <- mkTouch $ \x y ->
        code (realToFrac x) (realToFrac y)
    set ptr

setTouchesBegan :: (Double -> Double -> IO ()) -> IO ()
setTouchesBegan = setTouches_ setTouchesBegan_

setTouchesMoved :: (Double -> Double -> IO ()) -> IO ()
setTouchesMoved = setTouches_ setTouchesMoved_

setTouchesEnded :: (Double -> Double -> IO ()) -> IO ()
setTouchesEnded = setTouches_ setTouchesEnded_

applicationMain :: IO () -> IO () -> IO Int
applicationMain init dealloc = do
    fInit <- oneShot mkPlain init
    fDealloc <- oneShot mkPlain dealloc
    fromIntegral <$> applicationMain_ fInit fDealloc


newtype Sprite = Sprite CUInt

foreign import ccall safe "newSprite" newSprite_
    :: CString -> CUInt -> IO CUInt

newSprite :: String -> IO Sprite
newSprite filename = do
    withCStringLen filename $ \(cw_filename, filename_len) ->
        Sprite <$> newSprite_ cw_filename (fromIntegral filename_len)

foreign import ccall safe "drawSprite" drawSprite_
    :: CUInt -> CDouble -> CDouble -> IO ()

drawSprite :: Sprite -> (Double, Double) -> IO ()
drawSprite (Sprite tex) (x, y) = do
    drawSprite_ tex (realToFrac x) (realToFrac y)

foreign import ccall safe "drawRect" drawRect_
    :: CDouble -> CDouble -> CDouble -> CDouble -> IO ()

drawRect :: LineSegment -> IO ()
drawRect ((x0,y0),(x1,y1)) =
    drawRect_ (realToFrac x0) (realToFrac y0) (realToFrac x1) (realToFrac y1)

foreign import ccall unsafe "setDraw" setDraw_
    :: FunPtr (IO ()) -> IO ()

setDraw :: IO () -> IO ()
setDraw code = do
    setDraw_ =<< mkPlain code
    
foreign import ccall unsafe "getWidth" getWidth_
    :: IO CDouble

foreign import ccall unsafe "getHeight" getHeight_
    :: IO CDouble

getWidth :: IO Double
getWidth = realToFrac <$> getWidth_

getHeight :: IO Double
getHeight = realToFrac <$> getHeight_


------ Geometry ---------------------------------------------------------------

type Point = (Double, Double)
newtype Vector = Vector (Double, Double) deriving (Eq,Show)
type Ray = (Point, Vector)
type LineSegment = (Point, Point)


vectorTo :: Point -> Point -> Vector
vectorTo (x0, y0) (x1, y1) = Vector (x1-x0, y1-y0)

distanceTo :: Point -> Point -> Double
a `distanceTo` b = magnitude $ a `vectorTo` b

normalize :: Vector -> Vector
normalize v@(Vector (x, y)) =
    if mag < 0.000001
        then Vector (0,0)
        else Vector (x / mag, y / mag)
  where
    mag = magnitude v

dotProduct :: Vector -> Double -> Vector
dotProduct (Vector (x,y)) scalar = Vector (x * scalar, y * scalar)

polar :: Vector -> (Double, Double)  -- Magnitude, phase
polar (Vector (x,y)) = Complex.polar $ x :+ y

magnitude :: Vector -> Double
magnitude = fst . polar

minVector :: Vector -> Vector -> Vector
minVector v1 v2 =
    if magnitude v1 < magnitude v2
        then v1
        else v2

addVector :: Point -> Vector -> Point
addVector (x,y) (Vector (dx, dy)) = (x+dx, y+dy)

-- Angular distance of v1 relative to v2 in radians between -pi and +pi.
angularDist :: Vector -> Vector -> Double
angularDist v1 v2 =
         if d <  -pi then d + pi * 2
    else if d >= pi  then d - pi * 2
    else d
  where
    (_, dir1) = polar v1
    (_, dir2) = polar v2
    d = dir1 - dir2

lineLineIntersection :: LineSegment -> LineSegment -> Point
lineLineIntersection ((x1,y1),(x2,y2)) ((x3,y3),(x4,y4)) =
    (
        (p1 * (x3 - x4) - (x1 - x2) * p2) / bottom,
        (p1 * (y3 - y4) - (y1 - y2) * p2) / bottom
    )
  where
    p1     = x1*y2 - y1*x2
    p2     = x3*y4 - y3*x4
    bottom = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)

-- True if a point assumed to be on the line of the line segment is actually
-- on the line segment.
onLineSegment :: Point -> LineSegment -> Bool
onLineSegment (x,y) ((x0,y0),(x1,y1)) =
    x `between` (x0,x1) && y `between` (y0,y1)
  where
    x `between` (x0,x1) =
        if abs (x1 - x0) < 0.000001 then True
        else
            if x0 < x1 then x >= x0 && x <= x1
                       else x >= x1 && x <= x0

-- True if a point assumed to be on the line of the ray is actually on the ray
onRay :: Point -> Ray -> Bool
onRay (x,y) ((rx0,ry0),Vector (rvx, rvy)) =
    x `onRay` (rx0, rvx) && y `onRay` (ry0, rvy)
  where
    x `onRay` (r0, rv) =
        if abs rv < 0.000001
            then True
            else signum (x - r0) == signum rv 

-- Returns the intersection point of a ray and a line segment.
rayLineSegmentIntersection :: Ray -> LineSegment -> Maybe Point
rayLineSegmentIntersection ray ls =
    let ray2ls ((x,y), Vector (vx, vy)) = ((x,y),(x+vx, y+vy)) 
        int = lineLineIntersection (ray2ls ray) ls
    in  if int `onLineSegment` ls && int `onRay` ray
            then Just int
            else Nothing


------ The Game ---------------------------------------------------------------

type Microseconds = Integer
type Speed = Double
type SheepID = Int

dogSpeed = 200 / 1000000 :: Speed
sheepSpeed1 = 40 / 1000000 :: Speed
sheepSpeed2 = 160 / 1000000 :: Speed

data Locality = Locality {
        locPos   :: Point,
        locTime  :: Microseconds
    }

data Desire = Desire {
        desire       :: Double,
        desireChosen :: Bool
    }
    deriving Show
nullDesire = Desire 0 False

data Option = Option {
        optionName     :: String,
        evaluateOption :: SheepID -> GameState -> [(String, Double)],
        chooseOption   :: SheepID -> GameState -> Vector
    }

data Sheep = Sheep {
        sheepID      :: SheepID,
        sheepLoc     :: Locality,
        sheepDesires :: [Desire],
        sheepMotion  :: Vector
    }

sheepMoving :: Sheep -> Bool
sheepMoving sheep = sheepMotion sheep /= Vector (0,0)

data Dog = Dog {
        dogLoc   :: Locality
    }

data GameState = GameState {
        dog      :: Dog,
        sheep    :: Array Int Sheep,
        fences   :: [LineSegment]
    }

initialGameState now width height sheepPoints = GameState {
        dog = Dog {
                dogLoc = Locality (width / 2, 20) now
            },
        sheep =
            let indexes = [1..length sheepPoints]
            in  A.listArray (1,length sheepPoints) $ flip map (zip indexes sheepPoints) $ \(idx, pos) ->
                    Sheep {
                            sheepID  = idx,
                            sheepLoc = Locality pos now,
                            sheepDesires = map (const nullDesire) sheepOptions,
                            sheepMotion = Vector (0,0)
                        },
        fences =
            let l = 10
                b = 10
                r = width-10
                t = height-25
                penl = width * 0.35
                penb = height * 0.65
            in  [
                ((l,b),(l,t)),
                ((l,t),(r,t)),
                ((r,t),(r,b)),
                ((r,b),(l,b)),
                ((penl,penb),(penl,t))
            ]
    }

getTime :: IO Microseconds
getTime = do
    (TOD sec pico) <- getClockTime
    return $
        (fromIntegral sec::Integer) * 1000000 +
        (fromIntegral pico::Integer) `div` 1000000

fenceIntersection :: [LineSegment] -> Ray -> Maybe Point
fenceIntersection fences ray@(origin, _) =
    listToMaybe $ sortBy (comparing (`distanceTo` origin)) ints
  where
    ints = catMaybes $ map (rayLineSegmentIntersection ray) fences

moveTowards :: Microseconds -> Point -> Speed -> Locality -> Locality
moveTowards now dest speed loc = loc {
        locPos  = locPos loc `addVector` actualVec,
        locTime = now
    }
  where
    destVec = locPos loc `vectorTo` dest
    speedVec = normalize destVec `dotProduct` (fromIntegral delta_t * speed)
    actualVec = destVec `minVector` speedVec  -- don't overshoot the target
    delta_t = now - locTime loc

updateDog :: Microseconds -> Point -> Dog -> Dog
updateDog now dogDest dog = dog {
        dogLoc = moveTowards now dogDest dogSpeed (dogLoc dog)
    }

-- The list of things that a sheep can do
sheepOptions :: [Option]
sheepOptions =
        -- start moving in various directions
        flip map [0..15] (\step ->
            let theta = (fromIntegral step / 16) * 2 * pi
            in  move $ Vector (sin theta, cos theta))
        -- stop moving
        ++ [stop]

-- How resistant a sheep is to change, where 1 means 'not resistant at all'
-- and numbers greater than 1 are more resistant.
resistanceToChange = 1.4 :: Double

updateDesires :: GameState -> Sheep -> Sheep
updateDesires state sheep =
    let evaluations :: [Double]
        evaluations = flip map sheepOptions $ \option ->
            combineOptions $ evaluateOption option (sheepID sheep) state
        -- Alter desires according to evaluations
        desires' = flip map (zip (sheepDesires sheep) evaluations) $ \(des,eval) ->
            des {
                desire = (((desire des) * 0.1) `min` 10 `max` (-10)) + eval
            }
        (desires'', motion') = choose (zip sheepOptions desires')
    in  sheep {
                sheepDesires = desires'',
                sheepMotion  = motion'
            }
  where
    combineOptions :: [(String, Double)] -> Double
    combineOptions = sum . map snd

    choose :: [(Option, Desire)] -> ([Desire], Vector)
    choose optDesires =
         let (_, bestIdx, motion) = foldl' ch (0, -1, Vector (0,0)) (zip optDesires [0..])
             -- update the desireChosen flag according to what was chosen
             desires = snd $ unzip optDesires
             desires' = flip map (zip desires [0..]) $ \(des, idx) ->
                     des {
                         desireChosen = idx == bestIdx
                     }
         in  (desires', motion)
       where
         ch :: (Double, Int, Vector) -> ((Option, Desire), Int) -> (Double, Int, Vector)
         ch choiceSoFar@(bestScore, bestIdx, motion) ((option, des), idx) =
                 -- Amplify the desire that's currently chosen due to resistance
                 -- to change
             let amplifiedDesire = if desireChosen des then desire des * resistanceToChange
                                                       else desire des
             in  if amplifiedDesire > bestScore
                     then (amplifiedDesire, idx, chooseOption option (sheepID sheep) state)
                     else choiceSoFar

-- The Rules of Being A Sheep
-- 1. Run away from dog
-- 2. Don't run into a fence
-- 3. Follow another sheep
-- 4. Head towards empty space

move :: Vector -> Option
move dirVec = Option {
        optionName = "move "++show dirVec,
        evaluateOption = \sheepID state ->
            let me = sheep state ! sheepID
            in  -- the various reasons for moving
                [("away from dog",     fst (evaluateAwayFromDog dirVec me (dog state)) * 0.7),
                 ("don't cross fence", evaluateFence (fences state) dirVec me),
                 ("follow sheep",      evaluateFollow (A.elems $ sheep state) dirVec me * 5),
                 ("like open space",   evaluateOpenSpace (fences state) dirVec me * 0.5)],
        chooseOption = \sheepID state ->
            let me = sheep state ! sheepID
                (_, proximity) = evaluateAwayFromDog dirVec me (dog state)
                speed = if proximity < urgentDistance then sheepSpeed2
                                                      else sheepSpeed1
            in  dirVec `dotProduct` speed
    }

squared :: Num a => a -> a
squared x = x*x

-- Sheep feels unsafe within this distance of the dog
urgentDistance = 60 :: Double
dontCareDistance = 220 :: Double

evaluateAwayFromDog :: Vector -> Sheep -> Dog -> (Double, Double)
evaluateAwayFromDog dirVec sheep dog =
    let dogVec = locPos (sheepLoc sheep) `vectorTo` locPos (dogLoc dog)
        dogProximity = magnitude dogVec
        dogAngularDist = dogVec `angularDist` dirVec
        -- Desire to move away from the dog
        directionalComp = (((abs dogAngularDist - pi/4) `max` (-pi/4)) / (pi/2))
            * (((dontCareDistance - dogProximity) / dontCareDistance) `max` 0)
        -- Urgency if the dog is near
        proximityComp = (((urgentDistance - dogProximity) / urgentDistance) `max` 0)
            * (directionalComp `max` 0)
    in  (directionalComp + proximityComp, dogProximity)

evaluateFollow :: [Sheep] -> Vector -> Sheep -> Double
evaluateFollow sheep dirVec me =
    case closestSheep sheepProximities of
        Just sheep ->
            let herPos = locPos $ sheepLoc sheep
                myPos  = locPos $ sheepLoc me
                herVec = myPos `vectorTo` herPos
                prox = magnitude herVec
                distComp = if prox < 15 then 0
                             -- Only a slight preference for close sheep over distant ones
                             else ((200 - prox) / 200) `max` 0 `min` 1
                -- direction of sheep's motion relative to direction of sheep from me
                angDist = sheepMotion sheep `angularDist` herVec
                -- The following rule:
                -- Sheep is moving away? OOOOh!! Can't resist that!!
                movingAwayComp = ((pi - abs angDist) - pi / 2) `max` 0 / (pi / 2)
            in  distComp + movingAwayComp * 2
        Nothing -> 0
  where
    otherSheep = filter (\sh -> sheepID sh /= sheepID me) sheep  -- exclude me
    sheepProximities :: [(Sheep, Double)]
    sheepProximities = catMaybes $ flip map otherSheep $ \sheep ->
        let vec = (locPos $ sheepLoc me) `vectorTo` (locPos $ sheepLoc sheep)
            prox = magnitude vec
            adist = vec `angularDist` dirVec
            directionalComp = ((((pi - abs adist) - pi / 3) `max` 0) / (pi/2))
            distanceComp = -prox
        in  if directionalComp > 0.001
                then Just (sheep, distanceComp * directionalComp)
                else Nothing
    -- This is really the closest sheep in a given direction
    closestSheep :: [(Sheep, Double)] -> Maybe Sheep
    closestSheep [] = Nothing
    closestSheep proxes = Just $ fst $ foldl1' best proxes
        where
            best one@(_, value1) two@(_, value2) =
                if value1 > value2 then one else two

-- The distance that a sheep wants to be from a fence
fenceDistance = 32 :: Double

-- Sheep don't like to cross fences
evaluateFence :: [LineSegment] -> Vector -> Sheep -> Double
evaluateFence fences dirVec sheep =
    let pos = locPos $ sheepLoc sheep
        ray = (pos, dirVec)
        fenceInt = fenceIntersection fences ray
    in  case fenceInt of
            Just int ->
                -- If we're close to the fence, then make a large desire not to
                -- go through it.
                let dist = pos `distanceTo` int
                in  -(((fenceDistance - dist) `max` 0) * 100 / fenceDistance)
            Nothing  -> 0

-- Sheep like to head towards open space
evaluateOpenSpace :: [LineSegment] -> Vector -> Sheep -> Double
evaluateOpenSpace fences dirVec sheep =
    let pos = locPos $ sheepLoc sheep
        ray = (pos, dirVec)
        fenceInt = fenceIntersection fences ray
    in  case fenceInt of
            Just int ->
                let dist = pos `distanceTo` int
                in  (dist / 100) `min` 1
            Nothing  -> 0

stop :: Option
stop = Option {
        optionName = "stop",
        evaluateOption = \sheepID state ->
            let me = sheep state ! sheepID
            in
                [],
        chooseOption = \sheepID state -> Vector (0,0)
    }

moveSheep :: Microseconds -> [LineSegment] -> Sheep -> Sheep
moveSheep now fences sheep = sheep {
        sheepLoc = moveMe (sheepMotion sheep) (sheepLoc sheep)
    }
  where
    oldPos = locPos $ sheepLoc sheep
    ray = (oldPos, sheepMotion sheep)
    fenceInt = fenceIntersection fences ray

    moveMe :: Vector -> Locality -> Locality
    moveMe motion loc = loc {
            locPos  = {-trace ("fenceInt="++show fenceInt) $-} case fenceInt of
                Just int -> if oldPos `distanceTo` desiredPos < (oldPos `distanceTo` int) - 25
                                   then desiredPos
                                   else oldPos  -- Bang into fence
                Nothing  -> desiredPos,
            locTime = now
        }
      where
        delta_t = now - locTime loc
        desiredPos = locPos loc `addVector` (motion `dotProduct` fromIntegral delta_t)

updateSheep :: Microseconds -> GameState -> Sheep -> Sheep
updateSheep now state =
    moveSheep now (fences state).
    updateDesires state

updateState :: Microseconds -> Point -> GameState -> GameState
updateState now dogDest = execState $ do
    modify $ \st -> st { dog = updateDog now dogDest (dog st) }
    modify $ \st -> st { sheep = fmap (updateSheep now st) (sheep st) }

initialize :: IO ()
initialize = do
    putStrLn "Haskell init"
    width <- getWidth
    height <- getHeight
    now <- getTime

    let noOfSheep = 6

    sheepPoints <- forM [1..noOfSheep] $ \idx -> do
        x <- randomRIO (20, width-20)
        y <- randomRIO (20, height-20)
        return (x,y)

    stateRef <- newIORef $ initialGameState now width height sheepPoints
    -- Initial dog destination is taken from initial dog position
    dogDestRef <- newIORef . locPos . dogLoc . dog =<< readIORef stateRef

    dogSprite <- newSprite "dog"
    sheepSprite <- newSprite "sheep"
    setTouchesBegan $ \x y -> writeIORef dogDestRef (x,y)
    setTouchesMoved $ \x y -> writeIORef dogDestRef (x,y)
    setTouchesEnded $ \x y -> writeIORef dogDestRef (x,y)
    setDraw $ do
        state <- readIORef stateRef
        
        -- draw the fences
        forM_ (fences state) $ \((x0,y0),(x1,y1)) ->
            if x0 == x1 then drawRect ((x0-1,y0),(x0+1,y1)) else
            if y0 == y1 then drawRect ((x0,y0-1),(x1,y0+1)) else
                return ()

        dogDest <- readIORef dogDestRef
        now <- getTime
        let state' = updateState now dogDest state
        {-forM_ (A.elems $ sheep state) $ \sheep -> do
            forM_ (zip sheepOptions (sheepDesires sheep)) $ \(option, des) -> do
                putStrLn $ optionName option
                putStrLn $ "  -> " ++ show (evaluateOption option (sheepID sheep) state)
                putStrLn $ "  -> " ++ show des-}
        writeIORef stateRef state'
        forM_ (A.elems $ sheep state') $ \sheep -> do
            drawSprite sheepSprite (locPos $ sheepLoc sheep)
        drawSprite dogSprite (locPos $ dogLoc $ dog state')

dealloc :: IO ()
dealloc = do
    putStrLn "Haskell dealloc"

main = do
    hPutStrLn stderr "Haskell start"
    applicationMain initialize dealloc
    return ()

